#include <stdlib.h>
#include <string.h>

#include "filehandle.h"
#include "dsunzip.h"

#define MIN(a, b) ((a) < (b) ? (a) : (b))

Archive::Archive(FILE* file, u32 archiveFileSize, u32 filecount) {
    this->file = file;
    this->archiveFileSize = archiveFileSize;
    this->filecount = filecount;
    
    filenames = new char*[filecount];
    offsets   = new u32[filecount];
    filesizes = new u32[filecount];         
}

Archive::~Archive() {    
    if (file) {
        fclose(file);
        file = NULL;
    }

    for (u32 n = 0; n < filecount; n++) {
        delete[] filenames[n];
    }
    delete[] filenames;    
    delete[] offsets;
    delete[] filesizes;
}

ArchiveEntryHandle* Archive::OpenFile(char* filename) {
    int index = GetFileIndex(filename);
    if (index >= 0) {
        ArchiveEntryHandle* fh = new ArchiveEntryHandle(this, filename, index);
        return fh;
    }
    return NULL;
}

int Archive::GetFileIndex(char* name) {
    for (u32 n = 0; n < filecount; n++) {
        int x = 0;
        while (filenames[n][x] != '\0' && name[x] != '\0') {
            if (filenames[n][x] != name[x]) {
                break;
            }
            x++;
        }
        if (filenames[n][x] == '\0' && name[x] == '\0') {
            return n;
        }
    }
    return -1;
}

int Archive::Read(void* out, u32 offset, u32 len) {
    fseek(file, offset, SEEK_SET);
    return fread(out, 1, MIN(archiveFileSize - offset, len), file);
}

int findCentralDirectory(FILE* file, int pos) {
    #define bufferL (1<<10)
    #define readMax (1<<14)
    
    unsigned char* buffer = new unsigned char[bufferL + 3];    
    memset(buffer + bufferL, 0, 3); //Init last bytes to zero
    
    int read = 0;
    int bufferOffset = pos;
    do {
        read += bufferL;
        bufferOffset -= bufferL;
        if (bufferOffset < 0) {
            //There are a few bytes left, but I can't be bothered to search them...
            delete[] buffer;
            return -1;
        }
        fseek(file, -read, SEEK_CUR);
        
        if (fread(buffer, 1, bufferL, file) != bufferL) {
            //Read error
            delete[] buffer;
            return -1;
        }
        
        //"central directory" starts with 0x06054b50
        for (int n = (bufferL+3)-1; n >= 3; n--) {
            if (buffer[n-3]==0x50 && buffer[n-2]==0x4B && buffer[n-1]==0x05 && buffer[n]==0x06) {
                //Header found
                delete[] buffer;
                return bufferOffset + (n-3);
            }
        }
        
        //Move the first three bytes to the end of the buffer, they could match
        //during the next iteration
        buffer[bufferL]   = buffer[0];
        buffer[bufferL+1] = buffer[1];
        buffer[bufferL+2] = buffer[2];
    } while (bufferOffset > 0 && read < readMax);
    
    delete[] buffer;
    return -1;
}

Archive* OpenZipFile(char* filename) {
    FILE* file = fopen(filename, "rb");
    if (!file) {
        return NULL;
    }
    
    setvbuf(file, NULL, _IONBF, 0);

    fseek(file, 0, SEEK_END);
    int fileSize = ftell(file);
    
    int headerPos = findCentralDirectory(file, fileSize);
    if (headerPos < 0) {
        fclose(file);
        return NULL;
    }
    
    //Skip to header and ignore a few bytes I'm not interested in.
    fseek(file, headerPos + 10, SEEK_SET);
    
    unsigned short filecount;
    fread(&filecount, 2, 1, file);

    fseek(file, 4, SEEK_CUR); //boring

    unsigned int offset;
    fread(&offset, 4, 1, file);    
    
    Archive* archive = new Archive(file, fileSize, filecount);

    //Read file headers
    unsigned short filenameLength, extraFieldLength, commentLength;
    fseek(file, offset, SEEK_SET);

    //Skip boring stuff
    #define INITIAL_SKIP 24
    fseek(file, INITIAL_SKIP, SEEK_CUR);
    for (int n = 0; n < filecount; n++) {
        fread(&(archive->filesizes[n]), 4, 1, file);
        fread(&filenameLength, 2, 1, file);
        fread(&extraFieldLength, 2, 1, file);
        fread(&commentLength, 2, 1, file);
        
        fseek(file, 8, SEEK_CUR); //Boring...

        //Offset
        fread(&(archive->offsets[n]), 4, 1, file);
        
        //Filename
        char* name = new char[filenameLength + 1];
        fread(name, 1, filenameLength, file);
        name[filenameLength] = '\0';
        archive->filenames[n] = name;
        
        //The current value of offset is the offset of the local file header, not
        //the file itself. Change the offset to point to the file.
        archive->offsets[n] += 30 + filenameLength + extraFieldLength;
        
        //Skip to next relevant data
        fseek(file, extraFieldLength + commentLength + INITIAL_SKIP, SEEK_CUR);
                
        //fprintf(stdout, "%s Offset=%d Length=%d\n", filenames[n], offsets[n], filesizes[n]);
    }

    return archive;
}

Archive* OpenArchive(char* filename) {
    char path[260];
    Archive* archive = NULL;
    
    sprintf(path, "%s.zip", filename);    
    archive = OpenZipFile(path);
    if (archive) {
        return archive;
    }    
    
    return NULL;
}
void CloseArchive(Archive* archive) {
    if (archive) {
        delete archive;
    }
}
