#include "savemenu.h"
#include "vnds.h"
#include "vnimage.h"
#include "as_lib9.h"

#define ROWS 3
#define COLS 3
#define BAR_H 10

const int NUM_PAGES = (SAVE_SLOTS + (SLOTS_PER_PAGE-1)) / SLOTS_PER_PAGE;
const int NUM_BUTTONS = 5;

SaveMenu::SaveMenu(VNDS* vnds) {
    this->vnds = vnds;

    inited = false;
}
SaveMenu::~SaveMenu() {    
    Destroy();
}

void SaveMenu::Init() {
    if (inited) {
        Destroy();
    }
    inited = true;

    page = 0;
    selected = -1;
    buttonPressed = -1;
    backgroundDirty = true;

    frame = 0;
    animation = 0;
    slideDirection = 0;
    popupMenu = -1;

    for (int n = 0; n < NUM_BUTTONS; n++) {
        buttons[n] = NULL;
    }
    for (int n = 0; n < 2; n++) {
        popupButtons[n] = NULL;
    }
    for (int n = 0; n < SAVE_SLOTS; n++) {
        saveSlotImages[n] = new u16[SAVE_IMG_W * SAVE_IMG_H];
    }

    //Load images
    memset(background, 0, 256 * 192 * sizeof(u16));
    memset(backgroundImage, 0, 256 * 192 * sizeof(u16));
    LoadImage("../../data/save_background.jpg", backgroundImage, 256, 192);
    LoadImage("../../data/save_menu_buttons.png", textureImage, 128, 128);
    BlockCopy(defaultSlotImage, SAVE_IMG_W, textureImage, 128, 0, 32, SAVE_IMG_W, SAVE_IMG_H);
    BlockCopy(emptySlotImage, SAVE_IMG_W, textureImage, 128, 0, 80, SAVE_IMG_W, SAVE_IMG_H);
    
    // colorkeeeeeey
    for(int i = 0; i < 128*128; ++i) {
        if (textureImage[i] == (RGB15(31, 0, 31) | BIT(15))) {
            textureImage[i] = 0;
        }
    }

    //Setup video mode
    u16* temp = new u16[256 * 192];
    videobuf_main = (u16*)BG_BMP_RAM(0);
    videobuf_sub = (u16*)BG_BMP_RAM_SUB(0);

	swiCopy(videobuf_main, temp, 256 * 192 / 2 | COPY_MODE_WORD);
    memset(videobuf_main, 0, 256 * 192 * sizeof(u16));
    memset(videobuf_sub, 0, 256 * 192 * sizeof(u16));
    swiWaitForVBlank();
    lcdSwap();
    videoSetMode(MODE_5_3D | DISPLAY_BG3_ACTIVE);
	swiCopy(temp, videobuf_sub, 256 * 192 / 2 | COPY_MODE_WORD);
    delete[] temp;

    //Setup 3D stuff
    init3D();

    u16* texdata;

    glGenTextures(SLOTS_PER_PAGE, saveSlotTextures);
    texdata = new u16[64 * 64];
    memset(texdata, 0, 64 * 64 * 2);
    for (int n = 0; n < SLOTS_PER_PAGE; n++) {
    	glBindTexture(0, saveSlotTextures[n]);
    	glTexImage2D(0, 0, GL_RGBA, TEXTURE_SIZE_64, TEXTURE_SIZE_64, 0, TEXGEN_TEXCOORD, (u8*)texdata);
    }
    delete[] texdata;

	glGenTextures(1, &buttonTexture);
    texdata = new u16[128 * 128];
    for (int iy = 0; iy < 128; iy++) {
        memcpy(texdata + iy * 128, textureImage + iy * 128, 128 * sizeof(u16));
    }
	glBindTexture(0, buttonTexture);
	glTexImage2D(0, 0, GL_RGBA, TEXTURE_SIZE_128, TEXTURE_SIZE_128, 0, TEXGEN_TEXCOORD, (u8*)texdata);
    delete[] texdata;

    //Load save files
    Reload();
    SetPage(0);
}

void SaveMenu::Destroy() {
    if (inited) {
        inited = false;

        DeleteButtons();
        glInit();

        for (int n = 0; n < SAVE_SLOTS; n++) {
            delete[] saveSlotImages[n];
            saveSlotImages[n] = NULL;
        }

        //Restore video mode
        u16* temp = new u16[256 * 192];
   		swiCopy(videobuf_sub, temp, 256 * 192 / 2 | COPY_MODE_WORD);
        memset(videobuf_main, 0, 256 * 192 * sizeof(u16));
        memset(videobuf_sub, 0, 256 * 192 * sizeof(u16));
        swiWaitForVBlank();
        lcdSwap();
   		swiCopy(temp, videobuf_main, 256 * 192 / 2 | COPY_MODE_WORD);
        delete[] temp;

        vnds->VideoInit();
    }
}

void SaveMenu::DeleteButtons() {
    for (int n = 0; n < NUM_BUTTONS; n++) {
        if (buttons[n]) {
            delete buttons[n];
            buttons[n] = NULL;
        }
    }
    for (int n = 0; n < 2; n++) {
        if (popupButtons[n]) {
            delete popupButtons[n];
            popupButtons[n] = NULL;
        }
    }
}

void SaveMenu::ReloadTextures() {  
    //Be very careful when changing this code. It's very very easy to break things.
    while (GFX_STATUS & (1 << 27)); //Wait until graphics engine not busy

    //I need to set this to allow texture updates
    u32 temp = vramSetMainBanks(VRAM_A_MAIN_BG_0x06000000, VRAM_B_LCD,
        VRAM_C_SUB_BG_0x06200000, VRAM_D_LCD);

    //Generate save slot textures
    u16* vram = VRAM_B;
    for (int n = 0; n < SLOTS_PER_PAGE; n++) {
        int t = n + pageMin;        

        if (t < SAVE_SLOTS) {
            for (int n = 0; n < SAVE_IMG_W * SAVE_IMG_H; n++) {
                if (!IsActive()) {
                    //Set to a flashy color to notice programming mistakes more easily
                    (*vram) = RGB15(31, 0, 31) | BIT(15);
                } else {
                    (*vram) = saveSlotImages[t][n];
                }            
                vram++;
            }
            for (int n = 0; n < 64 * 16; n++) {
                (*vram) = 0;
                vram++;
            }
        }
    }

	vramRestoreMainBanks(temp);
}

void SaveMenu::Reload() {
    DeleteButtons();

    FILE* file;
    char path[128];
    for (int n = 0; n < SAVE_SLOTS; n++) {
        saveSlotExists[n] = true;
        //Read date
        sprintf(path, "save/save%.2d.sav", n);
        file = fopen(path, "rb");
        if (!file) {
            memcpy(saveSlotImages[n], emptySlotImage, SAVE_IMG_W * SAVE_IMG_H * sizeof(u16));
            saveSlotLabels[n][0] = '\0';
            saveSlotExists[n] = false;
            continue;
        }
        fclose(file);

        XMLNode xMainNode = XMLNode::openFileHelper(path, "save");
        strncpy(saveSlotLabels[n], xMainNode.getChildNode("date").getText(), 32);

        //Read image
        sprintf(path, "save/save%.2d.img", n);
        file = fopen(path, "rb");
        if (!file) {
            memcpy(saveSlotImages[n], defaultSlotImage, SAVE_IMG_W * SAVE_IMG_H * sizeof(u16));
            continue;
        }

        fread(saveSlotImages[n], sizeof(u16), SAVE_IMG_W * SAVE_IMG_H, file);
        fclose(file);
    }

    ReloadTextures();

    buttons[0] = new Button(buttonTexture, Rect(10, 150, 32, 32), Rect(64, 96, 32, 32));
    buttons[1] = new Button(buttonTexture, Rect(256 - 32 - 10, 150, 32, 32), Rect(96, 96, 32, 32));
    buttons[2] = new Button(buttonTexture, Rect(64, 150, 32, 32), Rect(64, 32, 32, 32));
    buttons[3] = new Button(buttonTexture, Rect(96 + 16, 150, 32, 32), Rect(96, 32, 32, 32));
    buttons[4] = new Button(buttonTexture, Rect(160, 150, 32, 32), Rect(64, 64, 32, 32));

    popupButtons[0] = new Button(buttonTexture, Rect(112 - 64, 100, 64, 32), Rect(0, 0, 64, 32));
    popupButtons[1] = new Button(buttonTexture, Rect(144, 100, 64, 32), Rect(64, 0, 64, 32));
}

void SaveMenu::OnButtonPress(int n) {
    if (animation == 0) {
        switch (n) {
            case 0:
                frame = 0;
                animation = 1;
                slideDirection = -1;
                selected -= SLOTS_PER_PAGE;
                break;
            case 1:
                frame = 0;
                animation = 1;
                slideDirection = 1;
                selected += SLOTS_PER_PAGE;
                break;
            case 2:
                if (n == 2) strcpy(popupMessage, "Load File?");
            case 3:
                if (n == 3) strcpy(popupMessage, "Save Progress?");
            case 4:
                if (n == 4) strcpy(popupMessage, "Return to Title?");

                popupMenu = n;
                backgroundDirty = true;
                break;
        }
    }
}
void SaveMenu::OnPopupButtonPress(int n) {
    if (animation == 0 && n == 0) {
        if (popupMenu == 2) {
            if (selected >= 0 && selected < SAVE_SLOTS) {
                vnds->Load(selected);
                backgroundDirty = true;
                Destroy();
            }
        } else if (popupMenu == 3) {
            if (selected >= 0 && selected < SAVE_SLOTS) {
                vnds->Save(selected);
                Reload();
            }
        } else if (popupMenu == 4) {
            vnds->ReturnToMenu();
        }
    }
    popupMenu = -1;
    backgroundDirty = true;
}

void SaveMenu::SetPage(int p) {
    backgroundDirty = true;
    page = p;

    if (page < 0) page += NUM_PAGES;
    if (page >= NUM_PAGES) page -= NUM_PAGES;

    pageMin = page * SLOTS_PER_PAGE;
    pageMax = MIN(SAVE_SLOTS, (page+1) * SLOTS_PER_PAGE) - 1;

    if (selected < pageMin) selected = pageMin;
    if (selected > pageMax) selected = pageMax;

    ReloadTextures();
}

void SaveMenu::Update()
{
    int os = selected;
    frame++;
    buttonPressedFrame++;
    if (buttonPressed >= 0 && buttonPressedFrame >= 5) {
        if (popupMenu >= 0) {
            OnPopupButtonPress(buttonPressed);
        } else {
            if (buttonPressed == 2) {
                if (saveSlotExists[selected]) {
                    OnButtonPress(buttonPressed);
                }
            }
            else {
                OnButtonPress(buttonPressed);
            }
        }
        buttonPressed = -1;
        if (!IsActive()) {
            return;
        }
    }

    if (animation == 1) {
        if (frame >= 20) {
            animation = 2;
            frame = 0;

            SetPage(page + slideDirection);
        }
    } else if (animation == 2) {
        if (frame >= 20) {
            animation = 0;
            frame = 0;
            slideDirection = 0;
        }
    }

    int oldSelected = selected;
    int newSelected = selected;

    int held = keysHeld();
    int down = keysDown();

    if (held & KEY_TOUCH) {
        touchPosition touchXY = touchReadXY();

        if (popupMenu >= 0) {
            for (int n = 0; n < 2; n++) {
                if (popupButtons[n]->Contains(touchXY.px, touchXY.py)) {
                    if (buttonPressed != n) {
                        if (down & KEY_TOUCH) {
                            buttonPressed = n;
                            buttonPressedFrame = 0;
                        }
                        break;
                    }
                }
            }
        } else {
            int col = (touchXY.px * COLS / 256);
            int row = ((touchXY.py-BAR_H) * ROWS / 192);
            newSelected = pageMin + row * COLS + col;
            if (newSelected >= pageMin && newSelected <= pageMax) {
                selected = newSelected;
            }

            for (int n = 0; n < NUM_BUTTONS; n++) {
                if (buttons[n] != NULL && buttons[n]->Contains(touchXY.px, touchXY.py)) {
                    if (buttonPressed != n) {
                        if (down & KEY_TOUCH) {
                            buttonPressed = n;
                            buttonPressedFrame = 0;
                        }
                        break;
                    }
                }
            }
        }
    }

    if (animation == 0) {
        if (popupMenu < 0) {
            if (down & (KEY_L | KEY_A)) {
                if (saveSlotExists[selected])
                    OnButtonPress(2);
            } else if (down & KEY_R) {
                OnButtonPress(3);
            } else if (down & KEY_B) {
                Destroy();
            }
            if (!IsActive()) {
                return;
            }

            if (down & (KEY_UP | KEY_DOWN | KEY_LEFT | KEY_RIGHT)) {
                if (down & KEY_LEFT) {
                    if (newSelected % COLS == 0) {
                        newSelected = pageMin;
                    }
                    newSelected -= 1;                    
                }
                if (down & KEY_RIGHT) {
                    if ((newSelected+1) % COLS == 0) {
                        newSelected = pageMax;
                    }
                    newSelected += 1;
                }
                if (down & KEY_UP) newSelected -= COLS;
                if (down & KEY_DOWN) newSelected += COLS;

                if (newSelected < pageMin || newSelected > pageMax) {
                    frame = 0;
                    animation = 1;
                    slideDirection = (newSelected < pageMin ? -1 : 1);

                    if (newSelected < 0) newSelected = SAVE_SLOTS - 1;
                    if (newSelected >= SAVE_SLOTS) newSelected = 0;
                }

                selected = newSelected;
            }
        } else {
            if (down & KEY_A) {
                buttonPressed = 0;
                buttonPressedFrame = 0;
            } else if (down & KEY_B) {
                buttonPressed = 1;
                buttonPressedFrame = 0;
            }
        }

        if (selected != oldSelected) {
            frame = 0;
        }
    }
    
    /*/if (os != selected) {
        PlaySound("click.wav");
    }*/
}

void SaveMenu::DrawBackground(Text* text) {
    text->SetScreen(background);
    text->SetPixelSize(10);
    text->SetInvert(false);

    memcpy(background, backgroundImage, 256 * 192 * sizeof(u16));
    for (int n = 0; n <= pageMax - pageMin; n++) {
        int t = n + pageMin;
        int row = (n / COLS);
        int col = n % COLS;
        int x = col * 256 / COLS + (256 / COLS - SAVE_IMG_W) / 2;
        int y = row * 192 / ROWS + SAVE_IMG_PAD + BAR_H;

        if (popupMenu >= 0) {
            //Draw Image
            for (int iy = 0; iy < SAVE_IMG_H; iy++) {
                memcpy(background + (y + iy) * 256 + x,
                    saveSlotImages[t] + iy * SAVE_IMG_W,
                    MIN(SAVE_IMG_W, 256 - x) * sizeof(u16));
            }
        }

        //Draw Label
        if (col == 0 && row <= 1) {
            int s = 3;
            int t = (y+SAVE_IMG_H+3) * 256;
            for (int n = 0; n < 256 * BAR_H; n++) {
                background[t] = RGB15(((background[t]&31)>>s), (((background[t]>>5)&31)>>s),
                    (((background[t]>>10)&31)>>s)) | BIT(15);
                t++;
            }
        }
        text->SetPen(x - 4, y + SAVE_IMG_H + BAR_H + 1);
        text->PrintString(saveSlotLabels[t]);
    }

    text->SetInvert(true);
    for (int n = 0; n < 256 * BAR_H; n++) {
        int a = 31 - (n/256);
        background[n] = RGB15(a, a, a) | BIT(15);
    }

    char string[128];

    vnds->TodayDate(string);
    text->SetPen(8, BAR_H - 2);
    text->PrintString(string);

    sprintf(string, "Page %d/%d", page+1, NUM_PAGES);
    text->SetPen(207, BAR_H - 2);
    text->PrintString(string);

	if (popupMenu >= 0) {
        int s = 3;
        for (int n = 0; n < 256 * 192; n++) {
            background[n] = RGB15(((background[n]&31)>>s), (((background[n]>>5)&31)>>s),
                (((background[n]>>10)&31)>>s)) | BIT(15);
        }

        text->SetInvert(false);
        text->SetPixelSize(18);
        int textW = text->GetStringWidth(popupMessage);
        text->SetPen(128 - textW/2, 80);
        text->PrintString(popupMessage);
    }

  	swiCopy(background, videobuf_main, 256 * 192 / 2 | COPY_MODE_WORD);
}

void SaveMenu::Draw(Text* text) {
    Update();
    if (!IsActive()) {
        return;
    }

    if (backgroundDirty) {
        DrawBackground(text);
        backgroundDirty = false;
    }

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glColor3f(1, 1, 1);

	glPolyFmt(POLY_ALPHA(31) | POLY_CULL_NONE | POLY_FORMAT_LIGHT0);
	glTranslatef(0, 0, -1);
    glScalef(INVSCALE, INVSCALE, 1);

    for (int n = 0; popupMenu < 0 && n <= pageMax - pageMin; n++) {
        int t = n + pageMin;
        int c = (n % COLS);

        int x = c * 256 / COLS + (256 / COLS - SAVE_IMG_W) / 2;
        int y = (n / COLS) * 192 / ROWS + SAVE_IMG_PAD + BAR_H;

        if (animation == 1) {
            if (frame < 20) {
                x = x - slideDirection * 256 * frame / 20;
            } else {
                continue;
            }
        } else if (animation == 2) {
            if (frame < 20) {
                if (slideDirection > 0) {
                    x +=  256 - 256 * frame / 20;
                } else {
                    x += -256 + 256 * frame / 20;
                }
            }
        }

       	glPushMatrix();
    	glTranslate(x, y);

        int scale = 100;
        if (t == selected) {
            int f = frame%100;

            if (f < 25) {
                scale = 100 + f;
            }
            else if (f < 75){
                scale = 150 - f;
            }
            else {
                scale = f;
            }
            
            x = -SAVE_IMG_W/2 * (scale-100) / 100;
            y = -SAVE_IMG_H/2 * (scale-100) / 100;

            glTranslate(x, y);
        }

        if (scale != 100) {
            glScalef(scale / 100.0f, scale / 100.0f, 1);
        }

        //Draw Image
    	drawQuad(saveSlotTextures[n], 0, 0, SAVE_IMG_W, SAVE_IMG_H);

    	glPopMatrix(1);
    }

	//Draw Buttons
	if (popupMenu >= 0) {
        for (int n = 0; n < 2; n++) {
            popupButtons[n]->Draw(buttonPressed == n);
        }
    } else {
        for (int n = 0; n < NUM_BUTTONS; n++) {
            if (buttons[n] != NULL) {
                buttons[n]->Draw(buttonPressed == n);
            }
        }
    }

	glPopMatrix(1);
    glFlush(0);
}

/*void SaveMenu::PlaySound(char* p)
{
    char path[256];
    sprintf(path, "/vnds/data/sound/%s", p); // hardcode, or ../../../../../?

    FILE* file = fopen(path, "rb");
    if (file == NULL) {
        return;
    }
    
    int len;
        
    fseek(file, 0, SEEK_END);
    len = ftell(file);
    fseek(file, 0, SEEK_SET);
    fread(soundbuf, sizeof(char), len, file);
    
    AS_SoundDefaultPlay((u8*)soundbuf, len, 127, 64, 0, 0);
    
    fclose(file);
}*/


bool SaveMenu::IsActive() {
    return inited;
}
